import typing as t
from http import HTTPStatus

from fastapi import HTTPException
from fastapi import Request
from starlette.responses import Response
from typing_extensions import Self

from ..responses import error_response

DEFAULT_ERROR_CODE: int = HTTPStatus.INTERNAL_SERVER_ERROR
DEFAULT_ERROR_DESCRIPTION: str = "An unknown internal error has occurred while processing the request."


class ApplicationError(HTTPException):
    """Base exception for the application."""

    def __init__(
        self: Self,
        detail: str = "",
        *,
        status_code: int | str | HTTPStatus | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: t.Any,
    ) -> None:
        if detail == "":
            detail = DEFAULT_ERROR_DESCRIPTION

        if status_code is None:
            status_code = DEFAULT_ERROR_CODE

        self.detail = detail
        self.status_code = status_code
        self.parameters = kwargs

        super().__init__(status_code=status_code, detail=detail, headers=headers)


_T = t.TypeVar("_T", ApplicationError, Exception)


def handle_application_error(_: Request, exc: _T) -> Response:
    return error_response(
        {
            "code": exc.status_code,
            "detail": exc.detail,
            "parameters": exc.parameters,
        },
        status_code=exc.status_code,
    )



import sys
from io import StringIO
from logging import Logger
from logging import getLogger
from pathlib import Path
from typing import TextIO

from typing_extensions import Self

try:
    from alembic.command import current
    from alembic.command import upgrade
    from alembic.config import Config
    from alembic.script import ScriptDirectory

except ImportError:
    error_message: str = "Alembic is not installed. Run `pip install iii-api-helper[database]`"
    raise ImportError(error_message) from None

logger: Logger = getLogger(__name__)


class Alembic:
    __slots__ = ["buffer", "ini_file", "scripts", "sql_uri"]

    alembic_logger: Logger = getLogger("alembic.runtime.migration")

    def __init__(self: Self, ini_file: Path | str, script_location: Path | str, sql_uri: str) -> None:
        self.ini_file = Path(ini_file)
        self.scripts = Path(script_location)
        self.sql_uri = sql_uri

        self.buffer: TextIO | StringIO = sys.stdout

    @property
    def script_directory(self: Self) -> ScriptDirectory:
        return ScriptDirectory.from_config(self.get_config())

    @property
    def current(self: Self) -> str:
        # https://stackoverflow.com/a/61770854
        self.alembic_logger.disabled = True
        self.buffer = StringIO()

        current(self.get_config())
        _out: str = self.buffer.getvalue().strip()

        self.alembic_logger.disabled = False
        self.buffer.close()
        self.buffer = sys.stdout

        return _out[:12]

    @property
    def head(self: Self) -> str:
        return self.script_directory.get_current_head()

    @property
    def upgradeable(self: Self) -> bool:
        # noinspection PyTestUnpassedFixture
        return self.current != self.head

    def get_config(self: Self) -> Config:
        _config: Config = Config(self.ini_file, stdout=self.buffer)

        # Patch the config file for the correct database URI
        if _config.get_main_option("sqlalchemy.url") != self.sql_uri:
            _config.set_main_option("sqlalchemy.url", self.sql_uri)

        _config.set_main_option("script_location", f"{self.scripts}")

        return _config

    def upgrade(self: Self, version: str = "head") -> None:
        upgrade(self.get_config(), version)



from logging import Logger
from logging import getLogger
from typing import Any

logger: Logger = getLogger(__name__)


class Singleton:
    _instance: Any = None

    def __new__(cls, *args: Any, **kwargs: Any) -> Any:  # noqa: ARG003
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance



from collections.abc import Callable
from copy import deepcopy
from logging import Logger
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import uvicorn
from fastapi import FastAPI as FastAPIBase
from starlette.middleware import Middleware
from typing_extensions import Self
from uvicorn._types import ASGIApplication

from . import config
from .config import LoggerBuilder
from .errors import FastAPIError
from .errors.app_error import ApplicationError
from .errors.app_error import handle_application_error
from .middlewares import LogRequestMiddleware
from .sentry import sentry_init

logger: Logger = getLogger(__name__)


class FastAPI(FastAPIBase):
    def __init__(
        self: Self,
        *,
        base_folder: Path,
        **extra: Any,
    ) -> None:
        """

        Args:
            base_folder:
            **extra:
        """
        if not base_folder.exists():
            msg: str = "The base folder does not exist. Please provide a valid base folder."
            raise FastAPIError(msg)

        self.setup_logger(base_folder, extra.pop("log_builder", None), **extra)

        middlewares: list[Middleware] = extra.pop("middleware", [])
        middlewares.append(
            Middleware(LogRequestMiddleware),  # type: ignore
        )

        if config.is_debug() or extra.get("debug", False) is True:
            extra.pop("debug", None)
            super().__init__(
                debug=True,
                title=config.get("APP_NAME"),
                middleware=middlewares,
                **extra,
            )
        else:
            openapi_url: str | None = extra.pop("openapi_url", None)

            super().__init__(
                title=config.get("APP_NAME"),
                middleware=middlewares,
                openapi_url=openapi_url,
                **extra,
            )

        self.add_exception_handler(ApplicationError, handle_application_error)

    def setup_logger(
        self: Self,
        base_folder: Path,
        log_builder: LoggerBuilder = None,
        **extra: Any,
    ) -> None:
        if log_builder:
            _logger: LoggerBuilder = log_builder

        else:
            options: dict[str, Any] = {}
            copied_extra: dict[str, Any] = deepcopy(extra)

            for key in copied_extra:
                if key == "logging_base_folder":
                    continue

                if key.startswith("logging_"):
                    options[key.replace("logging_", "")] = extra.pop(key)

            _logger: LoggerBuilder = LoggerBuilder(base_folder, **options)

        _logger.setup()

        self.LOGGING_CONFIG: dict[str, Any] = _logger.config

    if TYPE_CHECKING:
        run = uvicorn.run

    else:

        def run(
            self: Self,
            host: str = "127.0.0.1",
            port: int = 8000,
            *,
            app: ASGIApplication | Callable[..., Any] | str | None = None,
            show_swagger: bool = False,
            **kwargs: Any,
        ) -> None:
            error_message: str = ""

            if show_swagger and self.openapi_url is None:
                if self.title is None:
                    error_message = "A title must be provided for OpenAPI, e.g.: 'My API'"
                if self.version is None:
                    error_message = "A version must be provided for OpenAPI, e.g.: '2.1.0'"
                self.openapi_url = "/openapi.json"

                if error_message:
                    raise FastAPIError(error_message)

                self.setup()

            should_reload: bool = kwargs.pop("reload", config.is_reload())

            if should_reload:
                if not isinstance(app, str):
                    error_message = (
                        "The 'reload' option can only be used with 'app' as a string path to the application."
                    )
                    raise FastAPIError(error_message)

                module_str, _, attrs_str = app.partition(":")  # type: str, str, str
                if not module_str or not attrs_str:
                    error_message = 'Import string "{import_str}" must be in format "<module>:<attribute>".'
                    raise FastAPIError(error_message.format(import_str=app))

            uvicorn.run(
                app or self,
                host=host,
                port=port,
                reload=should_reload,
                log_config=self.LOGGING_CONFIG,
                **kwargs,
            )

    setup_sentry = sentry_init



import arrow
import pytest


@pytest.fixture()
def utctime_mock(monkeypatch: pytest.MonkeyPatch) -> None:
    def mock_utcnow() -> arrow.Arrow:
        return arrow.get("2021-01-01T00:00:00+00:00")

    monkeypatch.setattr(arrow, "utcnow", mock_utcnow)
